using Autofac;
using FitConnect;
using FitConnect.Exceptions;
using FitConnect.Interfaces;
using FitConnect.Models;
using FitConnect.Models.Api.Set;
using FitConnect.Models.v1.Api;
using FluentAssertions;
using Microsoft.Extensions.Logging;
using MockContainer;

namespace E2ETest;

[Order(20)]
[TestFixture]
public class AutoRejectReplyChannel : EndToEndTestBase {
    private SentSubmission _submission = null!;
    const string REPLY_CHANNEL_EMAIL = "dummy@trashmail.com";
    const string ELSTER_ACCOUNT_ID = "1234567890";

    [Order(10)]
    [Test]
    public void SendingSubmission() {
        var outgoing = SendableSubmission.Builder()
            .SetDestination(Settings.DestinationId)
            .SetServiceType(Settings.ServiceIdentifier, "Straight forward test")
            .SetJsonData(@"{""data"":""value""}", new Uri(JSON_SCHEMA_URI))
            .AddAttachments(
                Attachment.FromPath("Test.pdf",
                    "application/pdf", description: "Attachment #1", fileName: "Test.pdf"),
                Attachment.FromPath("Test.pdf", "application/pdf", description: "Attachment #2",
                    fileName: "Test.pdf"))
            .AddAttachments(Attachment.FromPath("Test.pdf",
                System.Net.Mime.MediaTypeNames.Application.Pdf, "Attachment #3",
                "Test.pdf"))
            .SetReplyChannel(new Elster(ELSTER_ACCOUNT_ID) {
                    Geschaeftszeichen = "ABC",
                    LieferTicket = "23453"
                }
            )
            .Build();

        _submission = Sender.SendAsync(outgoing).Result;

        _submission.Should().NotBeNull();


        _submission.AuthenticationTags!.Metadata.Should().NotBeNullOrWhiteSpace();
        _submission.AuthenticationTags.Data.Should().NotBeNullOrWhiteSpace();
        _submission.AuthenticationTags.Attachments!.All(item =>
                !(string.IsNullOrWhiteSpace(item.Value) ||
                  string.IsNullOrWhiteSpace($"{item.Key}")))
            .Should().BeTrue();
    }

    [Order(11)]
    [Test]
    public void SendingSubmission_WithoutAttachments() {
        var outgoing = SendableSubmission.Builder().SetDestination(Settings.DestinationId)
            .SetServiceType(Settings.ServiceIdentifier, "Straight forward test")
            .SetJsonData(@"{""data"":""value""}", new Uri(JSON_SCHEMA_URI))
            .Build();

        var submission = Sender.SendAsync(outgoing).Result;

        var caseId = submission.CaseId!;
        var submissionId = submission.SubmissionId!;
    }

    [Test]
    [Order(20)]
    public void Sender_GetSubmissionState() {
        // Act
        var status = Sender.GetStatusForSubmissionAsync(_submission)
            .Result;

        // Assert
        status.SubmissionState.Should().Be(EventType.Submit);
    }

    [Test]
    [Order(21)]
    public void Sender_GetEventLogAfterSubmit() {
        // Act
        var status = (Sender as Sender)!
            .GetEventLogAsync(_submission.CaseId, Settings.DestinationId).Result;

        // Assert
        status.Count.Should().BeGreaterThan(0);
        status.ForEach(
            s => Logger.LogInformation("Status {When} {Event}", s.EventTime, s.EventType));
    }

    [Test]
    [Order(30)]
    public void Subscriber_GetSubmissionState() {
        // Act
        var status = (Subscriber as Subscriber)!
            .GetEventLogAsync(_submission.CaseId, Settings.DestinationId).Result;

        // Assert
        status.Count.Should().BeGreaterThan(0);
        status.ForEach(
            s => Logger.LogInformation("Status {When} {Event}", s.EventTime, s.EventType));
    }

    [Test]
    [Order(40)]
    public void RequestSubmission() {
        var exception = Assert.Throws<AggregateException>(() => {
            var receivedSubmission =
                Subscriber.RequestSubmissionAsync(_submission.SubmissionId).Result;
        });
        exception!.InnerException!.InnerException.Should()
            .BeOfType(typeof(SecurityEventException));

        exception.InnerException.InnerException!.Message
            .Should().Contain("Reply channel is not supported");
    }

    [Test]
    [Order(50)]
    public void Sender_GetEventLog_AfterAccepting() {
        // Act
        var status = (Sender as Sender)!
            .GetEventLogAsync(_submission.CaseId, Settings.DestinationId).Result;

        // Assert
        status.Count.Should().BeGreaterThan(0);
        status.Any(s => s.EventType == EventType.Accept);

        var submitStatus = status.Single(s => s.EventType == EventType.Submit);
        submitStatus.Events?.SubmitSubmissionEvent?.AuthenticationTags?.Attachments.Should()
            .NotBeNullOrEmpty();

        status.ForEach(
            s => Logger.LogInformation("Status {When} {Event}", s.EventTime, s.EventType));
    }

    [Test]
    [Order(51)]
    public void Sender_GetSubmissionState_AfterAccepting() {
        var status = Sender.GetStatusForSubmissionAsync(_submission).Result;

        status.SubmissionState.Should().Be(EventType.Reject, "Submission should be rejected");
        status.Problems?.Where(p => p.Title == "Reply channel is not supported").Should()
            .NotBeNullOrEmpty();
    }

    [Test]
    [Order(60)]
    public void Sender_GetStatus_AfterAccepting() {
        // Act
        var (status, problems) =
            Sender.GetStatusForSubmissionAsync(_submission).Result;

        // Assert
        status.Should().Be(EventType.Reject, "Submission should be rejected");
        problems?.Where(p => p.Title == "Reply channel is not supported").Should()
            .NotBeNullOrEmpty();
    }


    [Test]
    [Order(70)]
    public void SendingSubmissionWithReplyChannel_ToDestinationWithoutReplyChannel_ThrowsError() {
        var container = Container.Create("NoReplyChannel");
        var settings = container.Resolve<IFitConnectSettings>();
        var sender = ClientFactory.GetSenderClient(FitConnectEnvironment.Test,
            settings.SenderClientId, settings.SenderClientSecret);
        var outgoing = SendableSubmission.Builder()
            .SetDestination(settings.DestinationId)
            .SetServiceType(settings.ServiceIdentifier, "Name")
            .SetJsonData(JSON_DATA,
                settings.JsonSchemaUri)
            .AddAttachments(Array.Empty<Attachment>())
            .SetReplyChannel(new EMail(REPLY_CHANNEL_EMAIL))
            .Build();

        var sent = sender.SendAsync(outgoing).Result;
        sent.Should().NotBeNull();

        sender.GetStatusForSubmissionAsync(sent).Result.SubmissionState.Should()
            .Be(EventType.Submit);

        var subscriber = ClientFactory.GetSubscriberClient(FitConnectEnvironment.Test,
            settings.SubscriberClientId, settings.SubscriberClientSecret,
            settings.PrivateDecryptionKeys, settings.PrivateSigningKey);


        var exception = Assert.Throws<AggregateException>(() => {
            var received = subscriber.RequestSubmissionAsync(sent.SubmissionId).Result;
            Console.WriteLine(received.GetData());
        });

        exception!.InnerExceptions.Select(e => e.InnerException).Any(e =>
            e!.Message.Contains("Reply channel")).Should().BeTrue();

        var submissionStatus = sender.GetStatusForSubmissionAsync(sent).Result;
        submissionStatus.SubmissionState.Should().Be(EventType.Reject);
        submissionStatus.Problems!.Any(p => p.Title == Problems.TitleUnsupportedReplyChannel)
            .Should().BeTrue();
    }
}
