using Autofac;
using FitConnect;
using FitConnect.Interfaces;
using FitConnect.Interfaces.Subscriber;
using FitConnect.Models;
using FluentAssertions;
using Microsoft.Extensions.Logging;
using Xunit.Abstractions;
using Xunit.Sdk;

namespace ValidationTests;

public class E2ETestsWithSchemaValidation {
    private readonly ISender _sender;
    private readonly ISubscriber _subscriber;
    private readonly ILogger _logger;
    private readonly IFitConnectSettings _settings;
    private ITestOutputHelper _output;

    public E2ETestsWithSchemaValidation() {
        _output = new TestOutputHelper();
        _logger = LoggerFactory
            .Create(c => {
                c.SetMinimumLevel(LogLevel.Trace);
                c.AddConsole();
            })
            .CreateLogger("Test");

        _settings = MockContainer.Container.Create().Resolve<IFitConnectSettings>();
        _sender = ClientFactory.GetSenderClient(FitConnectEnvironment.Test,
            _settings.SenderClientId, _settings.SenderClientSecret,
            _logger);

        _subscriber = ClientFactory.GetSubscriberClient(FitConnectEnvironment.Test,
            _settings.SubscriberClientId, _settings.SubscriberClientSecret,
            _settings.PrivateDecryptionKeys,
            _settings.PrivateSigningKey,
            _logger);
    }

    [Fact]
    public async void Transmission_WithoutSchemaValidation_ShouldPass() {
        _sender.SkipDataValidation = true;
        _subscriber.SkipDataValidation = true;

        var submission = SendableSubmission.Builder()
            .SetDestination(_settings.DestinationId)
            .SetServiceType(_settings.ServiceIdentifier, "Simple Transmission")
            .SetJsonData(@"{""name"":""John Doe"", ""age"":42}",
                _settings.JsonSchemaUri)
            .Build();

        var sentSubmission = await _sender.SendAsync(submission);

        (await _sender.GetStatusForSubmissionAsync(sentSubmission)).SubmissionState
            .Should().BeOneOf(new List<EventType>() { EventType.Submit, EventType.Notify });

        var onlineSubmissions =
            await _subscriber.GetAvailableSubmissionsAsync(_settings.DestinationId);
        foreach (var submissionForPickup in onlineSubmissions.Where(s =>
                     s.SubmissionId == sentSubmission.SubmissionId)) {
            var pickedUpSubmission = await _subscriber.RequestSubmissionAsync(submissionForPickup);
            pickedUpSubmission.Should().NotBeNull();
            pickedUpSubmission!.GetData().Should()
                .Be(@"{""name"":""John Doe"", ""age"":42}");

            await pickedUpSubmission.AcceptSubmissionAsync();
            _logger.LogInformation("Accepted submission");
        }

        (await _sender.GetStatusForSubmissionAsync(sentSubmission)).SubmissionState.Should()
            .Be(EventType.Accept);
    }

    [Fact]
    public async void Transmission_WithSchemaValidation_ShouldPass() {
        _sender.SkipDataValidation.Should().BeFalse();
        _subscriber.SkipDataValidation.Should().BeFalse();

        var submission = SendableSubmission.Builder()
            .SetDestination(_settings.DestinationId)
            .SetServiceType(_settings.ServiceIdentifier, "Simple Transmission")
            .SetJsonData(@"{""name"":""John Doe"", ""age"":42}",
                _settings.JsonSchemaUri)
            .Build();

        var sentSubmission = await _sender.SendAsync(submission);

        (await _sender.GetStatusForSubmissionAsync(sentSubmission)).SubmissionState
            .Should().BeOneOf(new List<EventType>() { EventType.Submit, EventType.Notify });

        var onlineSubmissions =
            await _subscriber.GetAvailableSubmissionsAsync(_settings.DestinationId);
        foreach (var submissionForPickup in onlineSubmissions.Where(s =>
                     s.SubmissionId == sentSubmission.SubmissionId)) {
            var pickedUpSubmission = await _subscriber.RequestSubmissionAsync(submissionForPickup);
            pickedUpSubmission.Should().NotBeNull();
            pickedUpSubmission!.GetData().Should()
                .Be(@"{""name"":""John Doe"", ""age"":42}");

            await pickedUpSubmission.AcceptSubmissionAsync();
            _logger.LogInformation("Accepted submission");
        }

        (await _sender.GetStatusForSubmissionAsync(sentSubmission)).SubmissionState.Should()
            .Be(EventType.Accept);
    }

    [Fact]
    public async void Transmission_WithSchemaValidation_ShouldFail() {
        _sender.SkipDataValidation.Should().BeFalse();
        _subscriber.SkipDataValidation.Should().BeFalse();

        var submission = SendableSubmission.Builder()
            .SetDestination(_settings.DestinationId)
            .SetServiceType(_settings.ServiceIdentifier, "Simple Transmission")
            .SetJsonData(@"{""name"":""John Doe"", ""age"":42}",
                _settings.JsonSchemaUri)
            .Build();

        var sentSubmission = await _sender.SendAsync(submission);

        (await _sender.GetStatusForSubmissionAsync(sentSubmission)).SubmissionState
            .Should().BeOneOf(new List<EventType>() { EventType.Submit, EventType.Notify });

        var onlineSubmissions =
            await _subscriber.GetAvailableSubmissionsAsync(_settings.DestinationId);
        foreach (var submissionForPickup in onlineSubmissions.Where(s =>
                     s.SubmissionId == sentSubmission.SubmissionId)) {
            await Assert.ThrowsAsync<AggregateException>(async () => {
                await _subscriber.RequestSubmissionAsync(submissionForPickup);
            });
        }

        (await _sender.GetStatusForSubmissionAsync(sentSubmission)).SubmissionState.Should()
            .Be(EventType.Reject);
    }
}
